/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  innerStab.cpp
///  \brief Implements the stabilizer of the nodes of our backtrack tree.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   06.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "innerStab.hpp"
#include "actingElement.hpp"

void InnerStab<Field, LINEAR>::minimizeIndependentK0(Matrix<Field>& matrix,
        ActingElement<Field, LINEAR>* const transporter)
{
    if (transporter == NULL)
    {
        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, NULL);
    }
    else
    {
        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, &transporter->m_A);
    }
    ++m_rank;
}

void InnerStab<Field, LINEAR>::minimize(unsigned int dummyVar, bool isInSpan, Matrix<Field>& matrix, ActingElement<
        Field, LINEAR>* const transporter)
{
    if (!isInSpan)
    {
        minimizeIndependentK0(matrix, transporter);
    }
    else
    {
        minimizeDependentK0(matrix, transporter);
    }
}

void InnerStab<Field, LINEAR>::minimizeDependentK0(Matrix<Field>& matrix, ActingElement<Field, LINEAR>* const transporter)
{
    if (transporter == NULL)
    {
        matrix.minimizeByColMult(false, m_rank, m_row_partition, m_fixedCols, NULL, NULL);
    }
    else
    {
        matrix.minimizeByColMult(false, m_rank, m_row_partition, m_fixedCols, &transporter->m_A, &transporter->m_phi);
    }
}

void InnerStab<Field, PLUSMINUSMONOMIAL>::minimize(unsigned int dummyVar, bool isInSpan, Matrix<Field>& matrix, ActingElement<
        Field, PLUSMINUSMONOMIAL>* const transporter)
{
    if (!isInSpan)
    {
        minimizeIndependentK0(matrix, transporter);
    }
    else
    {
        minimizeDependentK0(matrix, transporter);
    }
}

void InnerStab<Field, PLUSMINUSMONOMIAL>::minimizeDependentK0(Matrix<Field>& matrix, ActingElement<Field, PLUSMINUSMONOMIAL>* const transporter)
{
    if (transporter == NULL)
    {
        matrix.minimizeByColMult(true, m_rank, m_row_partition, m_fixedCols, NULL, NULL);
    }
    else
    {
        matrix.minimizeByColMult(true, m_rank, m_row_partition, m_fixedCols, &transporter->m_A, &transporter->m_phi);
    }
}

My_Long_Int InnerStab<Field, LINEAR>::getAutom(const Matrix<Field>& matrix,
        std::vector<ActingElement<Field, LINEAR> >& gens) const
{
    unsigned int n = matrix.getCols();
    unsigned int k = matrix.getRows();

    const std::vector<Vec<Field> >& matEntries = matrix.getEntries();
    std::vector<unsigned int> unit_pos;
    for (unsigned int i = 0; i < n; ++i)
    {
        unit_pos.push_back(matEntries[i].getFirstUnit().first);
    }

    std::vector < std::vector<unsigned int> > sets = m_row_partition.getSets();
    std::vector < Element<Field> > gensMult = Field::multGenerator();

    My_Long_Int result(1);

    for (unsigned int setIndex = 0; setIndex < sets.size(); ++setIndex)
    {
        result *= Field::my_multiplicativeGroupSize();
        const std::vector<unsigned int>& this_set = sets[setIndex];
        for (unsigned int i = 0; i < gensMult.size(); ++i)
        {
            Matrix < Field > A;
            A.identity(k);
            Vec < Field > phi(n, 1);

            for (unsigned int index = 0; index < this_set.size(); ++index)
            {
                unsigned int my_pos = this_set[index];
                A.setElement(my_pos, my_pos, gensMult[i]);
            }

            std::vector<unsigned int>::const_iterator this_setIterBeg = this_set.begin();
            std::vector<unsigned int>::const_iterator this_setIterEnd = this_set.end();
            Element < Field > multInverse = gensMult[i].multInverse();
            for (unsigned int i = 0; i < n; ++i)
            {
                std::vector<unsigned int>::const_iterator meetIter =
                        find(this_setIterBeg, this_setIterEnd, unit_pos[i]);
                if (meetIter != this_setIterEnd)
                {
                    phi.set(i, multInverse);
                }
            }
            gens.push_back(ActingElement<Field, LINEAR> (A, phi));
        }
    }

    return result;
}

My_Long_Int InnerStab<Field, PLUSMINUSMONOMIAL>::getAutom(const Matrix<Field>& matrix,
        std::vector<ActingElement<Field, PLUSMINUSMONOMIAL> >& gens) const
{
    unsigned int n = matrix.getCols();
    unsigned int k = matrix.getRows();

    const std::vector<Vec<Field> >& matEntries = matrix.getEntries();
    std::vector<unsigned int> unit_pos;
    for (unsigned int i = 0; i < n; ++i)
    {
        unit_pos.push_back(matEntries[i].getFirstUnit().first);
    }

    std::vector < std::vector<unsigned int> > sets = m_row_partition.getSets();
    My_Long_Int result(1);

    Element<Field> minusOne( -Element<Field>(1));

    for (unsigned int setIndex = 0; setIndex < sets.size(); ++setIndex)
    {
        result *= 2;
        const std::vector<unsigned int>& this_set = sets[setIndex];
        Matrix < Field > A;
        A.identity(k);
        Vec < Field > phi(n, 1);

        for (unsigned int index = 0; index < this_set.size(); ++index)
        {
            unsigned int my_pos = this_set[index];
            A.setElement(my_pos, my_pos, minusOne);
        }

        std::vector<unsigned int>::const_iterator this_setIterBeg = this_set.begin();
        std::vector<unsigned int>::const_iterator this_setIterEnd = this_set.end();

        for (unsigned int i = 0; i < n; ++i)
        {
            std::vector<unsigned int>::const_iterator meetIter =
                    find(this_setIterBeg, this_setIterEnd, unit_pos[i]);
            if (meetIter != this_setIterEnd)
            {
                phi.set(i, minusOne);
            }
        }
        gens.push_back(ActingElement<Field, PLUSMINUSMONOMIAL> (A, phi));
    }

    return result;
}

void InnerStab<Field, LINEAR>::fillActingElement(unsigned int k0, const Matrix<Field>& root,
        ActingElement<Field, LINEAR>& g, const std::vector<unsigned int>& to_fixSeq)
{
    unsigned int n = root.getCols();
    Matrix < Field > copy(root);
    copy.permuteCols(g.m_pi);
    std::vector<unsigned int>::const_iterator tofixIter = to_fixSeq.begin();
    for (unsigned int i = 0; i < n; ++i, ++tofixIter)
    {
        m_fixedCols.push_back(*tofixIter);
        m_isFixedCol[*tofixIter] = true;
        unsigned int height = (copy.getColumn(*tofixIter).height(m_rank)).first;
        minimize(k0, (height == Field::chainlength()), copy, &g);
    }
}

void InnerStab<Field, PLUSMINUSMONOMIAL>::fillActingElement(unsigned int k0, const Matrix<Field>& root,
        ActingElement<Field, PLUSMINUSMONOMIAL>& g, const std::vector<unsigned int>& to_fixSeq)
{
    unsigned int n = root.getCols();
    Matrix < Field > copy(root);
    copy.permuteCols(g.m_pi);
    std::vector<unsigned int>::const_iterator tofixIter = to_fixSeq.begin();

    for (unsigned int i = 0; i < n; ++i, ++tofixIter)
    {
        m_fixedCols.push_back(*tofixIter);
        m_isFixedCol[*tofixIter] = true;
        unsigned int height = (copy.getColumn(*tofixIter).height(m_rank)).first;
        minimize(k0, (height == Field::chainlength()), copy, &g);
    }
}

void InnerStab<Field, SEMILINEAR>::minimize(unsigned int dummyVar, bool isInSpan, Matrix<Field>& matrix, ActingElement<Field,
        SEMILINEAR>* const transporter)
{
    if (!isInSpan)

    {
        minimizeIndependentK0(matrix, transporter);
    }
    else
    {
        minimizeDependentK0(matrix, transporter);
    }
}

void InnerStab<Field, SEMILINEAR>::minimizeIndependentK0(Matrix<Field>& matrix, ActingElement<Field, SEMILINEAR>* const transporter)
{
    InnerStab<Field, LINEAR>::minimizeIndependentK0(matrix, transporter);
}

void InnerStab<Field, SEMILINEAR>::minimizeDependentK0(Matrix<Field>& matrix, ActingElement<Field, SEMILINEAR>* const transporter)
{
    InnerStab<Field, LINEAR>::minimizeDependentK0(matrix, transporter);
    if (m_frob_pow == 0)
    {
        return;
    }
    // we can minimize with the help of the automorphism
    unsigned int exp = 0;
    unsigned int act_col = m_fixedCols.back();
    const Vec<Field>& column(matrix.getColumn(act_col));
    Vec<Field>::const_reverse_iterator columnIter(column.rbegin() + (matrix.getRows() - m_rank) + 1); // the leading coefficient is equal to 1, can`t be minimized
    Vec<Field>::const_reverse_iterator columnIterEnd(column.rend());
    for (; columnIter < columnIterEnd && m_frob_pow != 0; ++columnIter)
    {
        Element<Field> actEntry = *columnIter;
        actEntry.applyFrobenius(exp);
        exp += actEntry.minimizeByFrobenius(m_frob_pow);
    }
    m_isFixedCol[act_col] = false; // now it will be minimized by the automorphisms.
    matrix.applyFrobenius(exp, &m_isFixedCol);
    m_isFixedCol[act_col] = true;
    if (transporter != NULL)
    {
        transporter->m_phi.applyFrobenius(exp);
        transporter->m_A.applyFrobenius(exp);
        transporter->m_frob_pow += exp;
    }
}

My_Long_Int InnerStab<Field, SEMILINEAR>::getAutom(const Matrix<Field>& matrix, std::vector<ActingElement<Field, SEMILINEAR> >& gens) const
{
    std::vector<ActingElement<Field, LINEAR> > linear_gens;
    My_Long_Int order = InnerStab<Field, LINEAR>::getAutom(matrix, linear_gens);
    gens.resize(gens.size() + linear_gens.size());
    std::copy(linear_gens.begin(), linear_gens.end(), gens.rbegin());
    if (m_frob_pow != 0)
    {
        gens.push_back(ActingElement<Field, SEMILINEAR> (matrix.getRows(), matrix.getCols(), m_frob_pow));
        order *= Field::get_r() / m_frob_pow;
    }
    return order;
}

void InnerStab<Field, SEMILINEAR>::fillActingElement(unsigned int k0, const Matrix<Field>& root,
        ActingElement<Field, SEMILINEAR>& g, const std::vector<unsigned int>& to_fixSeq)
{
    unsigned int n = root.getCols();
    Matrix<Field> copy(root);
    copy.permuteCols(g.m_pi);
    std::vector<unsigned int>::const_iterator tofixIter = to_fixSeq.begin();
    for (unsigned int i = 0; i < n; ++i, ++tofixIter)
    {
        m_fixedCols.push_back(*tofixIter);
        m_isFixedCol[*tofixIter] = true;
        unsigned int height = (copy.getColumn(*tofixIter).height(m_rank)).first;
        minimize(k0, (height == Field::chainlength()), copy, &g);
    }
}

