/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  node.cpp
///  \brief Specialization for Node
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   12.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "node.hpp"
#include "z4.hpp"
#include "f2_f2.hpp"

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<Field>(Node<Field, PERMUTATIONAL>& node, bool remove,
        std::vector<uchar>& res)
        {
    unsigned int n(node.m_matrix.getCols());
    res = std::vector<uchar>(n, 0);
    std::vector<uchar>::iterator resIterBeg(res.begin());
    std::vector<uchar>::iterator resIterEnd(res.end());

    std::vector<unsigned int> posToBlock(n);
    std::vector<unsigned int>::const_iterator posToBlockIterBeg(posToBlock.begin());
    std::vector<unsigned int>::const_iterator posToBlockIterEnd(posToBlock.end());
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree(), blockcounter = 0;
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    for (unsigned int col = 0; col < n; ++col)
    {
        if (col > 0 && node.m_blocks[col])
        {
            ++blockcounter;
        }
        posToBlock[inv_perm.apply(col, deg)] = blockcounter;
    }
    ++blockcounter;

    std::vector<DenseVec<Field> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<Field> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));

    std::vector<unsigned int>::const_iterator m_fixedUnElimIterBeg(node.m_fixedUnEliminated.begin());
    std::vector<unsigned int>::const_iterator m_fixedUnElimIterEnd(node.m_fixedUnEliminated.end());

    std::vector<Element<Field> > nontrivialUnits(Field::nontrivalUnits());
    std::vector<Element<Field> >::const_iterator nontrivialUnitsIterBeg(nontrivialUnits.begin());
    std::vector<Element<Field> >::const_iterator nontrivialUnitsIterEnd(nontrivialUnits.end());

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.
        if (remove && wordIter->operator[](lastFixedPreImg).isZero())
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            --node.m_minWord_workSize;
            continue;
        }
        bool hasNonzeroFixedUnEliminated(false);
        std::vector<unsigned int>::const_iterator m_fixedUnElimIter(m_fixedUnElimIterBeg);
        for (; m_fixedUnElimIter < m_fixedUnElimIterEnd; ++m_fixedUnElimIter)
        {
            const Element<Field>& entry(wordIter->operator[](*m_fixedUnElimIter));
            if (entry.isUnit())
            {
                hasNonzeroFixedUnEliminated = true;
                *wordIter *= entry.multInverse();
                break;
            }
        }
        addHashOfCodeword(*wordIter, resIterBeg, resIterEnd, posToBlockIterBeg, posToBlockIterEnd, blockcounter);
        if (!hasNonzeroFixedUnEliminated)
        {
            std::vector<Element<Field> >::const_iterator nontrivialUnitsIter(nontrivialUnitsIterBeg);
            for (; nontrivialUnitsIter < nontrivialUnitsIterEnd; nontrivialUnitsIter++)
            {
                addHashOfCodeword(*wordIter * (*nontrivialUnitsIter), resIterBeg, resIterEnd, posToBlockIterBeg,
                        posToBlockIterEnd, blockcounter);
            }
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
        }

template<>
void RefineMethods<PERMUTATIONAL>::addHashOfCodeword<Field>(const DenseVec<Field>& c,
        std::vector<uchar>::iterator& resIterBeg, std::vector<uchar>::iterator& resIterEnd, const std::vector<
        unsigned int>::const_iterator& posToBlockIterBeg,
        const std::vector<unsigned int>::const_iterator& posToBlockIterEnd, unsigned int blockcounter)
        {
    std::vector<std::vector<unsigned int> > loc_res(blockcounter, std::vector<unsigned int>(Field::my_size(), 0));
    std::vector<unsigned int>::const_iterator posToBlockIter(posToBlockIterBeg);
    DenseVec<Field>::const_iterator entryIter(c.begin());

    for (; posToBlockIter < posToBlockIterEnd; ++entryIter, ++posToBlockIter)
    {
        ++loc_res[*posToBlockIter][entryIter->getIndex()];
    }

    uchar myHash = HashGenerator::getHash(loc_res);
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin() + Field::my_size() * blockcounter);

    entryIter = c.begin();
    std::vector<uchar>::iterator resIter(resIterBeg);
    for (; resIter < resIterEnd; ++resIter, ++entryIter)
    {
        *resIter += myHash + *(hashIter + entryIter->getIndex());
    }
        }

template<>
template<>
std::vector<uchar>& RefineMethods<LINEAR>::minWordRefine<Z4>(Node<Z4, LINEAR>& node, bool remove,
        std::vector<uchar>& res)
        {
    unsigned int n(node.m_matrix.getCols());
    res = std::vector<uchar>(n, 0);

    std::vector<unsigned int> posToBlock(n);
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree();
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    std::vector < DenseVec<Z4> > masks;
    DenseVec < Z4 > *act_mask(NULL);
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterBeg(node.m_fixedUnEliminated.begin());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterEnd(node.m_fixedUnEliminated.end());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIter;

    for (unsigned int col = 0; col < n; ++col)
    {
        unsigned int preimage(inv_perm.apply(col, deg));
        if (node.m_blocks[col])
        {
            m_fixedUnEliminatedIter = find(m_fixedUnEliminatedIterBeg, m_fixedUnEliminatedIterEnd, preimage);
            if (m_fixedUnEliminatedIter != m_fixedUnEliminatedIterEnd || node.m_eliminatedPos[preimage])
            {
                continue;
            }
            else
            {
                masks.push_back(DenseVec<Z4> (n));
                act_mask = &masks.back();
            }
        }
        act_mask->setBits(preimage);
        //        std::cout<<*act_mask;
    }

    std::vector<unsigned int> fixedUnElimDatapos(node.m_fixedUnEliminated.size());
    std::vector<unsigned int> fixedUnElimBitRightShift(node.m_fixedUnEliminated.size());
    std::vector<unsigned int>::const_iterator iterBitRightShiftBegin(fixedUnElimBitRightShift.begin());
    std::vector<unsigned int>::const_iterator iterBitRightShiftEnd(fixedUnElimBitRightShift.end());
    std::vector<unsigned int>::const_iterator iterDataPosBegin(fixedUnElimDatapos.begin());

    bool existsFixedUnEliminated(!node.m_fixedUnEliminated.empty());
    if (existsFixedUnEliminated)
    {
        std::vector<unsigned int>::iterator iterBitRightShift(fixedUnElimBitRightShift.begin());
        std::vector<unsigned int>::iterator iterDataPos(fixedUnElimDatapos.begin());
        m_fixedUnEliminatedIter = m_fixedUnEliminatedIterBeg;

        for (; iterBitRightShift < iterBitRightShiftEnd; ++m_fixedUnEliminatedIter, ++iterBitRightShift, ++iterDataPos)
        {
            *iterDataPos = (*m_fixedUnEliminatedIter) / DenseVec<Z4>::ENTRIES_PER_WORD;
            *iterBitRightShift = 3 * ((*m_fixedUnEliminatedIter) % DenseVec<Z4>::ENTRIES_PER_WORD);
        }
    }

    //   print(std::cout, masks, "") << std::endl;

    std::vector<DenseVec<Z4> >::const_iterator maskIterBeg(masks.begin());
    std::vector<DenseVec<Z4> >::const_iterator maskIterEnd(masks.end());
    std::vector<DenseVec<Z4> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<Z4> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));
    std::vector<uchar>::iterator resIterBeg(res.begin());
    std::vector<uchar>::iterator resIterEnd(res.end());

    std::vector<unsigned int> heightSeq(n);
    std::vector<unsigned int>::iterator heightSeqIterBeg(heightSeq.begin());
    std::vector<unsigned int>::iterator heightSeqIterEnd(heightSeq.end());

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.
        if (remove && wordIter->operator[](lastFixedPreImg) == 0)
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            node.m_minWord_workSize--;
            continue;
        }

        std::vector<DenseVec<Z4> >::const_iterator maskIter(maskIterBeg);
        uchar myHash(0);
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (existsFixedUnEliminated)
        {
            std::vector<unsigned int>::const_iterator iterBitRightShift(iterBitRightShiftBegin);
            std::vector<unsigned int>::const_iterator iterDataPos(iterDataPosBegin);

            for (; iterBitRightShift < iterBitRightShiftEnd; ++iterBitRightShift, ++iterDataPos, ++hashIter)
            {
                myHash += (*hashIter)
                                * (Element<Z4> ((wordIter->getData(*iterDataPos) >> (*iterBitRightShift)) & 3).height());
            }
        }

        for (; maskIter < maskIterEnd; ++maskIter)
        {
            SymWeight < 2u > w = wordIter->getWeight(*maskIter);
            myHash += w.qty[0] * (*hashIter);
            ++hashIter;
            myHash += w.qty[1] * (*hashIter);
            ++hashIter;
        }

        //        wordIter->getHeightSeq(heightSeqIterBeg, heightSeqIterEnd);
        //        std::vector<unsigned int>::const_iterator heightIter(heightSeqIterBeg);
        std::vector<uchar>::const_iterator dataIter(wordIter->begin());
        unsigned int pos(DenseVec<Z4>::ENTRIES_PER_WORD);
        uchar word(0);
        uchar hash0(*hashIter);
        uchar hash1(*(hashIter + 1));

        for (std::vector<uchar>::iterator resIter(resIterBeg); resIter < resIterEnd; ++resIter, pos++)
        {
            if (pos == DenseVec<Z4>::ENTRIES_PER_WORD)
            {
                word = *dataIter;
                dataIter++;
                pos = 0;
            }

            if (word & ((uchar) 1))
            {
                // a unit
                *resIter += myHash ^ hash0;
            }
            else if (word & ((uchar) 2))
            {
                // the nilpotent element
                *resIter += myHash ^ hash1;
            }
            word >>= 3;
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
        }

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<Z4>(Node<Z4, PERMUTATIONAL>& node, bool remove,
        std::vector<uchar>& res)
        {
    unsigned int n(node.m_matrix.getCols());
    res = std::vector<uchar>(n, 0);

    std::vector<unsigned int> posToBlock(n);
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree();
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    std::vector<DenseVec<Z4> > masks;
    DenseVec<Z4>* act_mask(NULL);

    //    std::cout<<node<<std::endl;
    //    help::print(std::cout, node.m_fixedUnEliminated, " ")<<std::endl;
    //    help::print(std::cout,node.m_eliminatedPos, " " )<<std::endl;
    //    std::cout<<"los ghets"<<std::endl;
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterBeg(node.m_fixedUnEliminated.begin());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterEnd(node.m_fixedUnEliminated.end());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIter;

    for (unsigned int col = 0; col < n; ++col)
    {
        unsigned int preimage(inv_perm.apply(col, deg));
        if (node.m_blocks[col])
        {
            m_fixedUnEliminatedIter = find(m_fixedUnEliminatedIterBeg, m_fixedUnEliminatedIterEnd, preimage);
            if (m_fixedUnEliminatedIter != m_fixedUnEliminatedIterEnd || node.m_eliminatedPos[preimage])
            {
                continue;
            }
            else
            {
                masks.push_back( DenseVec<Z4> (n));
                act_mask = &masks.back();
            }
        }
        act_mask->setBits(preimage);
    }

    std::vector<unsigned int> fixedUnElimDatapos(node.m_fixedUnEliminated.size());
    std::vector<unsigned int> fixedUnElimBitRightShift(node.m_fixedUnEliminated.size());
    std::vector<unsigned int>::const_iterator iterBitRightShiftBegin(fixedUnElimBitRightShift.begin());
    std::vector<unsigned int>::const_iterator iterBitRightShiftEnd(fixedUnElimBitRightShift.end());
    std::vector<unsigned int>::const_iterator iterDataPosBegin(fixedUnElimDatapos.begin());

    bool existsFixedUnEliminated(!node.m_fixedUnEliminated.empty());
    if (existsFixedUnEliminated)
    {
        std::vector<unsigned int>::iterator iterBitRightShift(fixedUnElimBitRightShift.begin());
        std::vector<unsigned int>::iterator iterDataPos(fixedUnElimDatapos.begin());
        m_fixedUnEliminatedIter = m_fixedUnEliminatedIterBeg;

        for (; iterBitRightShift < iterBitRightShiftEnd; ++m_fixedUnEliminatedIter, ++iterBitRightShift, ++iterDataPos)
        {
            *iterDataPos = (*m_fixedUnEliminatedIter) / DenseVec<Z4>::ENTRIES_PER_WORD;
            *iterBitRightShift = 3 * ((*m_fixedUnEliminatedIter) % DenseVec<Z4>::ENTRIES_PER_WORD);
        }
    }

    //   print(std::cout, masks, "") << std::endl;

    std::vector<DenseVec<Z4> >::const_iterator maskIterBeg(masks.begin());
    std::vector<DenseVec<Z4> >::const_iterator maskIterEnd(masks.end());
    std::vector<DenseVec<Z4> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<Z4> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));
    std::vector<uchar>::iterator resIterBeg(res.begin());
    std::vector<uchar>::iterator resIterEnd(res.end());

    //    std::vector<unsigned int> heightSeq( n );
    //    std::vector<unsigned int>::iterator heightSeqIterBeg(heightSeq.begin());
    //    std::vector<unsigned int>::iterator heightSeqIterEnd(heightSeq.end());

    const std::vector<uchar>::const_iterator hashIterBeg(HashGenerator::begin() + 4);
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
    uchar hash0(*hashIter);
    uchar hash1(*(hashIter + 1));
    uchar hash2(*hashIter + 2);
    uchar hash3(*(hashIter + 3));

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.
        if (remove && wordIter->operator[](lastFixedPreImg) == 0)
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            node.m_minWord_workSize--;
            continue;
        }

        hashIter = hashIterBeg;
        std::vector<DenseVec<Z4> >::const_iterator maskIter(maskIterBeg);
        uchar myHash1(0);
        uchar myHash3(0);
        if (existsFixedUnEliminated)
        {
            std::vector<unsigned int>::const_iterator iterBitRightShift(iterBitRightShiftBegin);
            std::vector<unsigned int>::const_iterator iterDataPos(iterDataPosBegin);

            for (; iterBitRightShift < iterBitRightShiftEnd; ++iterBitRightShift, ++iterDataPos, ++hashIter)
            {
                Element<Z4> entry((wordIter->getData(*iterDataPos) >> (*iterBitRightShift)) & 3);
                if (entry.isUnit())
                {
                    if (entry.isOne())
                    {
                        myHash1 += (*hashIter);
                        myHash3 += (*hashIter) * 3;
                    }
                    else
                    {
                        myHash1 += (*hashIter) * 3;
                        myHash3 += (*hashIter);
                    }
                }
                else
                {
                    myHash1 += (*hashIter) * (entry.getIndex());
                    myHash3 += (*hashIter) * (entry.getIndex());
                }
            }
        }

        for (; maskIter < maskIterEnd; ++maskIter)
        {
            std::vector<unsigned int> w(wordIter->getCompleteWeight(*maskIter));
            myHash1 += w[1] * (*hashIter);
            myHash3 += w[3] * (*hashIter);
            ++hashIter;
            myHash1 += w[2] * (*hashIter);
            myHash3 += w[2] * (*hashIter);
            ++hashIter;
        }

        std::vector<uchar>::const_iterator dataIter(wordIter->begin());
        unsigned int pos(DenseVec<Z4>::ENTRIES_PER_WORD);
        uchar word(0);
        for (std::vector<uchar>::iterator resIter(resIterBeg); resIter < resIterEnd; ++resIter, pos++)
        {
            if (pos == DenseVec<Z4>::ENTRIES_PER_WORD)
            {
                word = *dataIter;
                dataIter++;
                pos = 0;
            }
            switch (word & ((uchar) 3))
            {
            case 0:
                *resIter += myHash1 ^ hash0;
                *resIter += myHash3 ^ hash0;
                break;
            case 1:
                *resIter += myHash1 ^ hash1;
                *resIter += myHash3 ^ hash3;
                break;
            case 2:
                *resIter += myHash1 ^ hash2;
                *resIter += myHash3 ^ hash2;
                break;
            case 3:
                *resIter += myHash1 ^ hash3;
                *resIter += myHash3 ^ hash1;
                break;
            }
            word >>= 3;
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
        }

template<>
template<>
std::vector<uchar>& RefineMethods<LINEAR>::minWordRefine<F2_F2>(Node<F2_F2, LINEAR>& node, bool remove, std::vector<
        uchar>& res)
        {
    unsigned int n(node.getCols());
    res = std::vector<uchar>(n, 0);

    std::vector<unsigned int> posToBlock(n);
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree();
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    std::vector < DenseVec<F2_F2> > masks;
    DenseVec < F2_F2 > *act_mask(NULL);

    for (unsigned int col = 0; col < n; ++col)
    {
        unsigned int preimage(inv_perm.apply(col, deg));
        if (node.m_blocks[col])
        {
            if (node.m_eliminatedPos[preimage])
            {
                continue;
            }
            else
            {
                masks.push_back(DenseVec<F2_F2> (n));
                act_mask = &masks.back();
            }
        }
        act_mask->setBits(preimage);
    }

    std::vector<unsigned int> fixedUnElimDatapos(node.m_fixedUnEliminated.size());
    std::vector<unsigned int> fixedUnElimBitRightShift(node.m_fixedUnEliminated.size());
    std::vector<unsigned int>::const_iterator iterBitRightShiftBegin(fixedUnElimBitRightShift.begin());
    std::vector<unsigned int>::const_iterator iterBitRightShiftEnd(fixedUnElimBitRightShift.end());
    std::vector<unsigned int>::const_iterator iterDataPosBegin(fixedUnElimDatapos.begin());

    bool existsFixedUnEliminated(node.m_fixedUnEliminated.size() != 0);
    if (existsFixedUnEliminated)
    {
        std::vector<unsigned int>::const_iterator iter(node.m_fixedUnEliminated.begin());
        std::vector<unsigned int>::iterator iterBitRightShift(fixedUnElimBitRightShift.begin());
        std::vector<unsigned int>::iterator iterDataPos(fixedUnElimDatapos.begin());

        for (; iterBitRightShift < iterBitRightShiftEnd; ++iter, ++iterBitRightShift, ++iterDataPos)
        {
            *iterDataPos = (*iter) >> 5;
            *iterBitRightShift = 2 * ((*iter) & DenseVec<F2_F2>::BITPOSMASK);
        }
    }

    //   print(std::cout, masks, "") << std::endl;

    std::vector<DenseVec<F2_F2> >::const_iterator maskIterBeg(masks.begin());
    std::vector<DenseVec<F2_F2> >::const_iterator maskIterEnd(masks.end());
    std::vector<DenseVec<F2_F2> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<F2_F2> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));
    std::vector<uchar>::iterator resIterBeg(res.begin());
    std::vector<uchar>::iterator resIterEnd(res.end());

    std::vector<unsigned int> heightSeq(n);
    std::vector<unsigned int>::iterator heightSeqIterBeg(heightSeq.begin());
    std::vector<unsigned int>::iterator heightSeqIterEnd(heightSeq.end());

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.

        if (remove && wordIter->operator[](lastFixedPreImg) == 0)
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            node.m_minWord_workSize--;
            continue;
        }

        std::vector<DenseVec<F2_F2> >::const_iterator maskIter(maskIterBeg);
        uchar myHash(0);
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (existsFixedUnEliminated)
        {
            std::vector<unsigned int>::const_iterator iterBitRightShift(iterBitRightShiftBegin);
            std::vector<unsigned int>::const_iterator iterDataPos(iterDataPosBegin);

            for (; iterBitRightShift < iterBitRightShiftEnd; ++iterBitRightShift, ++iterDataPos, ++hashIter)
            {
                myHash += (*hashIter)
                                * (Element<F2_F2> ((wordIter->getData(*iterDataPos) >> (*iterBitRightShift)) & 3).height());
            }
        }

        for (; maskIter < maskIterEnd; ++maskIter)
        {
            SymWeight < 2u > w = wordIter->getWeight(*maskIter);
            myHash += w.qty[0] * (*hashIter);
            ++hashIter;
            myHash += w.qty[1] * (*hashIter);
            ++hashIter;
        }

        std::vector<uchar>::const_iterator dataIter(wordIter->begin());
        unsigned int pos(0);
        uchar word(0);
        uchar hash0(*hashIter);
        uchar hash1(*(hashIter + 1));

        for (std::vector<uchar>::iterator resIter(resIterBeg); resIter < resIterEnd; ++resIter, pos++)
        {
            if ((pos & DenseVec<F2_F2>::BITPOSMASK) == 0)
            {
                word = *dataIter;
                dataIter++;
                pos = 0;
            }

            if (word & ((uchar) 1))
            {
                // a unit
                *resIter += myHash ^ hash0;
            }
            else if (word & ((uchar) 2))
            {
                // the nilpotent element
                *resIter += myHash ^ hash1;
            }
            word >>= 2;
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
        }

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<F2_F2>(Node<F2_F2, PERMUTATIONAL>& node, bool remove,
        std::vector<uchar>& res)
        {
    unsigned int n(node.getCols());
    res = std::vector<uchar>(n, 0);

    std::vector<unsigned int> posToBlock(n);
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree();
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    std::vector<DenseVec<F2_F2> > masks;
    DenseVec<F2_F2>* act_mask(NULL);
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterBeg(node.m_fixedUnEliminated.begin());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterEnd(node.m_fixedUnEliminated.end());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIter;

    for (unsigned int col = 0; col < n; ++col)
    {
        unsigned int preimage(inv_perm.apply(col, deg));
        if (node.m_blocks[col])
        {
            m_fixedUnEliminatedIter = find(m_fixedUnEliminatedIterBeg, m_fixedUnEliminatedIterEnd, preimage);
            if (m_fixedUnEliminatedIter != m_fixedUnEliminatedIterEnd || node.m_eliminatedPos[preimage])
            {
                continue;
            }
            else
            {
                masks.push_back(DenseVec<F2_F2> (n));
                act_mask = &masks.back();
            }
        }
        act_mask->setBits(preimage);
    }

    std::vector<unsigned int> fixedUnElimDatapos(node.m_fixedUnEliminated.size());
    std::vector<unsigned int> fixedUnElimBitRightShift(node.m_fixedUnEliminated.size());
    std::vector<unsigned int>::const_iterator iterBitRightShiftBegin(fixedUnElimBitRightShift.begin());
    std::vector<unsigned int>::const_iterator iterBitRightShiftEnd(fixedUnElimBitRightShift.end());
    std::vector<unsigned int>::const_iterator iterDataPosBegin(fixedUnElimDatapos.begin());

    bool existsFixedUnEliminated( !node.m_fixedUnEliminated.empty());
    if (existsFixedUnEliminated)
    {
        std::vector<unsigned int>::iterator iterBitRightShift(fixedUnElimBitRightShift.begin());
        std::vector<unsigned int>::iterator iterDataPos(fixedUnElimDatapos.begin());
        m_fixedUnEliminatedIter = m_fixedUnEliminatedIterBeg;

        for (; iterBitRightShift < iterBitRightShiftEnd; ++m_fixedUnEliminatedIter, ++iterBitRightShift, ++iterDataPos)
        {
            *iterDataPos = (*m_fixedUnEliminatedIter) >> 5;
            *iterBitRightShift = 2 * ((*m_fixedUnEliminatedIter) & DenseVec<F2_F2>::BITPOSMASK);
        }
    }

    std::vector<DenseVec<F2_F2> >::const_iterator maskIterBeg(masks.begin());
    std::vector<DenseVec<F2_F2> >::const_iterator maskIterEnd(masks.end());
    std::vector<DenseVec<F2_F2> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<F2_F2> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));
    std::vector<uchar>::iterator resIterBeg(res.begin());
    std::vector<uchar>::iterator resIterEnd(res.end());

    const std::vector<uchar>::const_iterator hashIterBeg(HashGenerator::begin() + 4);
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
    uchar hash0(*hashIter);
    uchar hash1(*(hashIter + 1));
    uchar hash2(*hashIter + 2);
    uchar hash3(*(hashIter + 3));

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.

        if (remove && wordIter->operator[](lastFixedPreImg) == 0)
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            node.m_minWord_workSize--;
            continue;
        }

        hashIter = hashIterBeg;
        std::vector<DenseVec<F2_F2> >::const_iterator maskIter(maskIterBeg);
        uchar myHash1(0);
        uchar myHash3(0);
        if (existsFixedUnEliminated)
        {
            std::vector<unsigned int>::const_iterator iterBitRightShift(iterBitRightShiftBegin);
            std::vector<unsigned int>::const_iterator iterDataPos(iterDataPosBegin);

            for (; iterBitRightShift < iterBitRightShiftEnd; ++iterBitRightShift, ++iterDataPos, ++hashIter)
            {
                Element<F2_F2> entry((wordIter->getData(*iterDataPos) >> (*iterBitRightShift)) & 3);
                if (entry.isUnit())
                {
                    if (entry.isOne())
                    {
                        myHash1 += (*hashIter);
                        myHash3 += (*hashIter) * 3;
                    }
                    else
                    {
                        myHash1 += (*hashIter) * 3;
                        myHash3 += (*hashIter);
                    }
                }
                else
                {
                    myHash1 += (*hashIter) * (entry.getIndex());
                    myHash3 += (*hashIter) * (entry.getIndex());
                }
            }
        }

        for (; maskIter < maskIterEnd; ++maskIter)
        {
            std::vector<unsigned int> w(wordIter->getCompleteWeight(*maskIter));
            myHash1 += w[1] * (*hashIter);
            myHash3 += w[3] * (*hashIter);
            ++hashIter;
            myHash1 += w[2] * (*hashIter);
            myHash3 += w[2] * (*hashIter);
            ++hashIter;
        }

        std::vector<uchar>::const_iterator dataIter(wordIter->begin());
        unsigned int pos(0);
        uchar word(0);
        for (std::vector<uchar>::iterator resIter(resIterBeg); resIter < resIterEnd; ++resIter, pos++)
        {
            if ((pos & DenseVec<F2_F2>::BITPOSMASK) == 0)
            {
                word = *dataIter;
                dataIter++;
                pos = 0;
            }
            switch (word & ((uchar) 3))
            {
            case 0:
                *resIter += myHash1 ^ hash0;
                *resIter += myHash3 ^ hash0;
                break;
            case 1:
                *resIter += myHash1 ^ hash1;
                *resIter += myHash3 ^ hash3;
                break;
            case 2:
                *resIter += myHash1 ^ hash2;
                *resIter += myHash3 ^ hash2;
                break;
            case 3:
                *resIter += myHash1 ^ hash3;
                *resIter += myHash3 ^ hash1;
                break;
            }
            word >>= 2;
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
        }

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<BinaryField>(Node<BinaryField, PERMUTATIONAL>& node,
        bool remove, std::vector<uchar>& res)
        {
    //    std::cout<<*this<<std::endl;
    unsigned int n(node.m_matrix.getCols());

    res = std::vector<uchar>(n, 0);

    std::vector<unsigned int> posToBlock(n);
    Permutation inv_perm;
    //    std::cout<<m_perm_to_sol<<std::endl;

    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree();
    unsigned int lastFixedPreImg(0);
    unsigned int lastFixedPreImg_datapos(0);
    uchar lastFixedPreImg_bitmask(1);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
        lastFixedPreImg_datapos = lastFixedPreImg >> 6;
        lastFixedPreImg_bitmask = ((uchar) 1) << (lastFixedPreImg & DenseVec<BinaryField>::BITPOSMASK);
    }

    std::vector<DenseVec<BinaryField> > masks;
    DenseVec<BinaryField>* act_mask(NULL);

    //    std::vector<unsigned int> dummyOccur(n, 0);

    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterBeg(node.m_fixedUnEliminated.begin());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIterEnd(node.m_fixedUnEliminated.end());
    std::vector<unsigned int>::const_iterator m_fixedUnEliminatedIter;

    for (unsigned int col = 0; col < n; ++col)
    {
        unsigned int preimage(inv_perm.apply(col, deg));
        if (node.m_blocks[col])
        {
            m_fixedUnEliminatedIter = find(m_fixedUnEliminatedIterBeg, m_fixedUnEliminatedIterEnd, preimage);
            if (m_fixedUnEliminatedIter != m_fixedUnEliminatedIterEnd || node.m_eliminatedPos[preimage])
            {
                continue;
            }
            else
            {
                masks.push_back(DenseVec<BinaryField> (n));
                act_mask = &masks.back();
            }
        }
        act_mask->setBit(preimage);
        //        std::cout<<*act_mask<<std::endl;
    }

    std::vector<unsigned int> fixedUnElimDatapos(node.m_fixedUnEliminated.size());
    std::vector<uchar> fixedUnElimBitMask(node.m_fixedUnEliminated.size());
    std::vector<uchar>::const_iterator iterBitMaskBegin(fixedUnElimBitMask.begin());
    std::vector<uchar>::const_iterator iterBitMaskEnd(fixedUnElimBitMask.end());
    std::vector<unsigned int>::const_iterator iterDataPosBegin(fixedUnElimDatapos.begin());

    bool existsFixedUnEliminated(!node.m_fixedUnEliminated.empty());
    if (existsFixedUnEliminated)
    {
        std::vector<uchar>::iterator iterBitMask(fixedUnElimBitMask.begin());
        std::vector<unsigned int>::iterator iterDataPos(fixedUnElimDatapos.begin());
        m_fixedUnEliminatedIter = m_fixedUnEliminatedIterBeg;

        for (; iterBitMask < iterBitMaskEnd; ++m_fixedUnEliminatedIter, ++iterBitMask, ++iterDataPos)
        {
            *iterDataPos = (*m_fixedUnEliminatedIter) >> 6;
            *iterBitMask = ((uchar) 1) << ((*m_fixedUnEliminatedIter) & DenseVec<BinaryField>::BITPOSMASK);
        }
    }

    std::vector<DenseVec<BinaryField> >::const_iterator maskIterBeg(masks.begin());
    std::vector<DenseVec<BinaryField> >::const_iterator maskIterEnd(masks.end());
    std::vector<DenseVec<BinaryField> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<BinaryField> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin()
            + (node.m_minWords.size() - node.m_minWord_workSize));
    std::vector<uchar>::iterator resIterBeg(res.begin());
    //    std::vector<uchar>::iterator resIterEnd(res.end());

    //    std::vector<unsigned int> heightSeq( this->getCols() );
    //    std::vector<unsigned int>::iterator heightSeqIterBeg(heightSeq.begin());
    //    std::vector<unsigned int>::iterator heightSeqIterEnd(heightSeq.end());

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.
        if (remove && wordIter->getBits(lastFixedPreImg_datapos, lastFixedPreImg_bitmask) == 0)
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            --node.m_minWord_workSize;
            continue;
        }

        uchar myHash((uchar) 0);
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (existsFixedUnEliminated)
        {
            std::vector<uchar>::const_iterator iterBitMask(iterBitMaskBegin);
            std::vector<unsigned int>::const_iterator iterDataPos(iterDataPosBegin);

            unsigned int x(0);
            for (; iterBitMask < iterBitMaskEnd; ++iterBitMask, ++iterDataPos, ++hashIter)
            {
                if (wordIter->getBits(*iterDataPos, *iterBitMask) == 0)
                {
                    myHash += *hashIter;
                }
                ++x;
            }
        }

        std::vector<DenseVec<BinaryField> >::const_iterator maskIter(maskIterBeg);
        for (; maskIter < maskIterEnd; ++maskIter, ++hashIter)
        {
            myHash += (*hashIter) * (wordIter->getWeight(*maskIter).qty);
        }

        std::vector<uchar>::const_iterator wordDataIter(wordIter->begin());
        uchar word(0);
        unsigned int pos(0);
        for (std::vector<uchar>::iterator resIter(resIterBeg); pos < n; ++resIter, ++pos)
        {
            if ((pos & DenseVec<BinaryField>::BITPOSMASK) == 0)
            {
                word = *wordDataIter;
                ++wordDataIter;
            }
            //            dummyOccur[pos] += ((unsigned int) word)&1;

            if (((unsigned int) word) & 1)
            {
                *resIter += myHash ^ *hashIter;
            }
            else
            {
                *resIter += myHash;
            }
            word >>= 1;
        }
        ++wordIter;
        ++i;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    //    help::print(std::cout, dummyOccur, " ")<<std::endl;
    //    if (remove)
    //    {
    //        std::cout<<m_innerStab.m_fixedCols.back()<<std::endl;
    //    }
    //    std::cout<<m_innerStab.m_fixedCols.size()<<std::endl;
    //    help::print(std::cout, res, " ")<<std::endl;

    return res;
        }

template<>
template<>
std::vector<uchar>& RefineMethods<LINEAR>::supportRefine<Field>(const Node<Field, LINEAR>& node, const std::vector<
        unsigned int>& heights, std::vector<uchar>& res)
        {
    return supportRefineK0(node, heights, res);
        }

template<>
template<>
std::vector<uchar>& RefineMethods<SEMILINEAR>::supportRefine<Field>(const Node<Field, SEMILINEAR>& node,
        const std::vector<unsigned int>& heights, std::vector<uchar>& res)
        {
    unsigned int frob_pow(node.getInnerStab().m_frob_pow);
    if (frob_pow == 0)
    {
        return RefineMethods<LINEAR>::supportRefineK0((const Node<Field, LINEAR>&) node, heights, res);
    }

    res = std::vector<uchar>(node.getCols(), 0);
    std::vector<uchar>::iterator resIter(res.begin());
    std::vector<unsigned int>::const_iterator heightIter(heights.begin());

    const std::vector<Vec<Field> >& matrixCols(node.getMatrix().getEntries());
    std::vector<Vec<Field> >::const_iterator colIter(matrixCols.begin());
    std::vector<Vec<Field> >::const_iterator colIterEnd(matrixCols.end());

    std::vector<bool>::const_iterator isFixIter(node.getInnerStab().m_isFixedCol.begin());

    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
    unsigned int rank(node.getInnerStab().m_rank);
    const UnordPartition& row_partition(node.getInnerStab().m_row_partition);

    for (; colIter < colIterEnd; ++colIter, ++resIter, ++heightIter, ++isFixIter)
    {
        if (*isFixIter)
        {
            continue;
        }
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (*heightIter == 1)
        {
            std::vector < Element<Field> > quotient_crossed_rows(rank, 0);
            unsigned int applied_frob_pow(0);
            unsigned int stabilizer_frob(frob_pow);

            std::vector<Element<Field> >::const_iterator colEntIter(colIter->begin());
            std::vector<Element<Field> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            unsigned int row = 0;

            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter, ++row)
            {
                if (colEntIter->isUnit())
                {
                    int setIndex = row_partition.getSetIndex(row);
                    Element < Field > &quotient = quotient_crossed_rows[setIndex];
                    if (quotient.isZero())
                    {
                        quotient = colEntIter->multInverse();
                        *resIter += *hashIter;
                    }
                    else
                    {
                        Element < Field > a(quotient * (*colEntIter));
                        a.applyFrobenius(applied_frob_pow);
                        applied_frob_pow += a.minimizeByFrobenius(stabilizer_frob);
                        *resIter += (*hashIter) * a.getIndex();
                    }
                }
            }
        }
    }
    return res;
        }

template<>
std::vector<uchar>& RefineMethods<PLUSMINUSMONOMIAL>::supportRefine<Field>(const Node<Field, PLUSMINUSMONOMIAL>& node,
        const std::vector<unsigned int>& heights, std::vector<uchar>& res);

template<>
std::vector<uchar>& RefineMethods<PLUSMINUSMONOMIAL>::minWordRefine<Field>(Node<Field, PLUSMINUSMONOMIAL>& node, bool remove, std::vector<uchar>& res)
{
    unsigned int n(node.m_matrix.getCols());
    res = std::vector<uchar>(n, 0);
    std::vector<uchar>::iterator resIterBeg(res.begin());
    std::vector<uchar>::iterator resIterEnd(res.end());

    std::vector<unsigned int> posToBlock(n);
    std::vector<unsigned int>::const_iterator posToBlockIterBeg(posToBlock.begin());
    std::vector<unsigned int>::const_iterator posToBlockIterEnd(posToBlock.end());
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree(), blockcounter = 0;
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    for (unsigned int col = 0; col < n; ++col)
    {
        if (col > 0 && node.m_blocks[col])
        {
            ++blockcounter;
        }
        posToBlock[inv_perm.apply(col, deg)] = blockcounter;
    }
    ++blockcounter;

    std::vector<DenseVec<Field> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<Field> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));

    std::vector<unsigned int>::const_iterator m_fixedUnElimIterBeg(node.m_fixedUnEliminated.begin());
    std::vector<unsigned int>::const_iterator m_fixedUnElimIterEnd(node.m_fixedUnEliminated.end());

    std::vector<Element<Field> > nontrivialUnitsPlusMinus(Field::nontrivalUnitsPlusMinus());
    std::vector<Element<Field> >::const_iterator nontrivialUnitsPlusMinusIterBeg( nontrivialUnitsPlusMinus.begin() ), nontrivialUnitsPlusMinusIterEnd( nontrivialUnitsPlusMinus.end() );

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.
        if (remove && wordIter->operator[](lastFixedPreImg).isZero())
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            --node.m_minWord_workSize;
            continue;
        }
        bool hasNonzeroFixedUnEliminated(false);
        std::vector<unsigned int>::const_iterator m_fixedUnElimIter(m_fixedUnElimIterBeg);
        for (; m_fixedUnElimIter < m_fixedUnElimIterEnd; ++m_fixedUnElimIter)
        {
            const Element<Field>& entry(wordIter->operator[](*m_fixedUnElimIter));
            if (entry.isUnit())
            {
                hasNonzeroFixedUnEliminated = true;
                *wordIter *= entry.multInverse();
                break;
            }
        }
        addHashOfCodeword(*wordIter, resIterBeg, resIterEnd, posToBlockIterBeg, posToBlockIterEnd, blockcounter);
        if (!hasNonzeroFixedUnEliminated)
        {
            std::vector<Element<Field> >::const_iterator nontrivialUnitsIter(nontrivialUnitsPlusMinusIterBeg);
            for (; nontrivialUnitsIter < nontrivialUnitsPlusMinusIterEnd; nontrivialUnitsIter++)
            {
                addHashOfCodeword(*wordIter * (*nontrivialUnitsIter), resIterBeg, resIterEnd, posToBlockIterBeg,
                        posToBlockIterEnd, blockcounter);
            }
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
}

template<>
void RefineMethods<PLUSMINUSMONOMIAL>::addHashOfCodeword<Field>(const DenseVec<Field>& c,
        std::vector<uchar>::iterator& resIterBeg, std::vector<uchar>::iterator& resIterEnd, const std::vector<
        unsigned int>::const_iterator& posToBlockIterBeg,
        const std::vector<unsigned int>::const_iterator& posToBlockIterEnd, unsigned int blockcounter)
{
    std::vector<std::vector<unsigned int> > loc_res(blockcounter, std::vector<unsigned int>(Field::my_size(), 0));
    std::vector<unsigned int>::const_iterator posToBlockIter(posToBlockIterBeg);
    DenseVec<Field>::const_iterator entryIter(c.begin());
    Element<Field> negEntry;

    for (; posToBlockIter < posToBlockIterEnd; ++entryIter, ++posToBlockIter)
    {
        negEntry = -(*entryIter);
        if(negEntry < *entryIter)
        {
            ++loc_res[*posToBlockIter][negEntry.getIndex()];
        }
        else
        {
            ++loc_res[*posToBlockIter][entryIter->getIndex()];
        }
    }

    uchar myHash = HashGenerator::getHash(loc_res);
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin() + Field::my_size() * blockcounter);

    entryIter = c.begin();
    std::vector<uchar>::iterator resIter(resIterBeg);
    for (; resIter < resIterEnd; ++resIter, ++entryIter)
    {
        negEntry = -(*entryIter);
        if(negEntry < *entryIter)
        {
            *resIter += myHash + *(hashIter + negEntry.getIndex());
        }
        else
        {
            *resIter += myHash + *(hashIter + entryIter->getIndex());
        }
    }
}

template<>
std::vector<uchar>& RefineMethods<PLUSMINUSMONOMIAL>::supportRefine(const Node<Field, PLUSMINUSMONOMIAL>& node, const std::vector<
        unsigned int>& heights, std::vector<uchar>& res)
        {
    // only on s0 component...
    unsigned int n(node.getCols());
    res = std::vector<uchar>(n, 0);
    std::vector<uchar>::iterator resIter(res.begin());
    std::vector<unsigned int>::const_iterator heightIter(heights.begin());
    const std::vector<Vec<Field> >& matrixCols(node.getMatrix().getEntries());
    std::vector<Vec<Field> >::const_iterator colIter(matrixCols.begin());
    std::vector<Vec<Field> >::const_iterator colIterEnd(matrixCols.end());

    std::vector<bool>::const_iterator isFixIter(node.getInnerStab().m_isFixedCol.begin());
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());

    unsigned int rank(node.getInnerStab().m_rank);
    const UnordPartition& row_partition(node.getInnerStab().m_row_partition);

    for (; colIter < colIterEnd; ++colIter, ++resIter, ++heightIter, ++isFixIter)
    {
        if (*isFixIter)
        {
            continue;
        }
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (*heightIter == 1)
        {
            std::vector<Element<Field> > quotient_crossed_rows(rank, 0);
            std::vector<Element<Field> >::const_iterator colEntIter(colIter->begin());
            std::vector<Element<Field> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            unsigned int row = 0;

            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter, ++row)
            {
                if (colEntIter->isUnit())
                {
                    int setIndex = row_partition.getSetIndex(row);
                    Element<Field>& quotient = quotient_crossed_rows[setIndex];
                    if (quotient.isZero())
                    {
                        if(*colEntIter < (-*colEntIter) )
                        {
                            quotient = Element<Field>(1);
                            *resIter += *hashIter *  colEntIter->getIndex();
                        }
                        else
                        {
                            quotient = -Element<Field>(1);
                            *resIter += *hashIter * (-(*colEntIter)).getIndex();
                        }
                    }
                    else
                    {
                        *resIter += (*hashIter) * (quotient * (*colEntIter)).getIndex();
                    }
                }
            }
        }
    }
    return res;
        }

