/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  node.hpp
///  \brief The nodes of the backtrack tree.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef NODE_H_
#define NODE_H_

#include <set>
#include <map>

#include "actingElement.hpp"
#include "abstractCASModule.h"
#include "weights.hpp"
#include "codewordEnumerator.hpp"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \struct NextFixInfo
/// \brief   Gives the next block to fix, i.e. the columns [\a index, \f$\ldots\f$, \a index + \a blocksize -1] and tells you
/// if these columns are in the span of the fixed ones.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
struct NextFixInfo
{
    unsigned int m_index; ///<The index of an unfixed column.
    unsigned int m_blockSize; ///< The size of the block, containing the column \a index.
    bool m_isInSpan; ///< A flag if the column \a index is in the span of the fixed ones.

    NextFixInfo(unsigned int i, unsigned int size, bool inSpan) :
        m_index(i), m_blockSize(size), m_isInSpan(inSpan)
    {
    }
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class RefinementValues
/// \brief The results of the refinement procedures, for one level.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class RefinementValues
{
public:
    std::vector<unsigned int> m_heights; ///<The results for Matrix<RING>::heightSequence.
    std::vector<uchar> m_support; ///<The results for Matrix<RING>::supportSequence.

    /// \brief The results for Node<RING>::minWordRefine.
    ///
    /// This method is called iteratively, therefore we need a vector for the resuls.
    std::vector<std::vector<uchar> > m_minWordImg;
};

template<typename RING, AlgTypeE ALGTYPE>
class GenNode;

template<typename RING, AlgTypeE ALGTYPE>
class Node;

template<AlgTypeE ALGTYPE>
class RefineMethods
{
public:
    template<typename RING>
    static std::vector<uchar>& minWordRefine(Node<RING, ALGTYPE>& node, bool remove, std::vector<uchar>& res);

    template<typename RING>
    static std::vector<uchar>& supportRefine(const Node<RING, ALGTYPE>& node, const std::vector<unsigned int>& heights,
            std::vector<uchar>& res);

    template<typename RING>
    static std::vector<uchar>& supportRefineK0(const Node<RING, ALGTYPE>& node,
            const std::vector<unsigned int>& heights, std::vector<uchar>& res);

    template<typename RING>
    static std::vector<uchar>& supportRefineK1(const Node<RING, ALGTYPE>& node,
            const std::vector<unsigned int>& heights, std::vector<uchar>& res);
};

template<>
class RefineMethods<PLUSMINUSMONOMIAL>
{
private:
    template<typename RING>
    static void addHashOfCodeword(const DenseVec<RING>& c,
            std::vector<uchar>::iterator& resIterBeg, std::vector<uchar>::iterator& resIterEnd,
            const std::vector<unsigned int>::const_iterator& posToBlockIterBeg,
            const std::vector<unsigned int>::const_iterator& posToBlockIterEnd, unsigned int blockcounter);

public:
    template<typename RING>
    static std::vector<uchar>& minWordRefine(Node<RING, PLUSMINUSMONOMIAL>& node, bool remove, std::vector<uchar>& res);

    template<typename RING>
    static std::vector<uchar>& supportRefine(const Node<RING, PLUSMINUSMONOMIAL>& node, const std::vector<unsigned int>& heights,
            std::vector<uchar>& res);
};

template<>
class RefineMethods<PERMUTATIONAL>
{
private:
    template<typename RING>
    static void addHashOfCodeword(const DenseVec<RING>& c,
            std::vector<uchar>::iterator& resIterBeg, std::vector<uchar>::iterator& resIterEnd,
            const std::vector<unsigned int>::const_iterator& posToBlockIterBeg,
            const std::vector<unsigned int>::const_iterator& posToBlockIterEnd, unsigned int blockcounter);
public:
    template<typename RING>
    static std::vector<uchar>& minWordRefine(Node<RING, PERMUTATIONAL>& node, bool remove, std::vector<uchar>& res);

    template<typename RING>
    static std::vector<uchar>& supportRefine(const Node<RING, PERMUTATIONAL>& node,
            const std::vector<unsigned int>& heights, std::vector<uchar>& res);
};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class Node
/// \brief The nodes of the backtrack tree through the symmetric group.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename RING, AlgTypeE ALGTYPE>
class Node
{
private:
    static const unsigned int MINIMAL_WORKSIZE = 500; ///< There should be at least this number of words in the invariant set.
    static const unsigned int MAXIMAL_INIT_WORKSIZE = 1024; ///< There should be at least this number of words in the invariant set we calculate for the root node.
    static const unsigned int MAXIMAL_NR_OF_WEIGHTS = 5; ///< We keep this maximal number of different weights for the invariant set.

//    static const unsigned int MINIMAL_WORKSIZE = 30; ///< There should be at least this number of words in the invariant set.
//    static const unsigned int MAXIMAL_INIT_WORKSIZE = 50; ///< There should be at least this number of words in the invariant set we calculate for the root node.
//    static const unsigned int MAXIMAL_NR_OF_WEIGHTS = 5; ///< We keep this maximal number of different weights for the invariant set.

    Matrix<RING> m_matrix; ///< The generator matrix.
    OrdPartition m_blocks; ///< The partition of the columns.
    InnerStab<RING, ALGTYPE> m_innerStab; ///< The inner stabilizer.
    Permutation m_perm_to_sol; ///< The last fully performed permutation on the matrix.

    std::vector<DenseVec<RING> > m_minWords; ///< An invariant subset of codewords.
    unsigned long m_minWord_workSize; ///< The first \a m_minWord_workSize words form another invariant subset for the actual \a lvl.
    std::vector<bool> m_eliminatedPos; ///< The positions which were used to reduce the invariant subset. These don't give any information anymore.
    std::vector<unsigned int> m_fixedUnEliminated; ///< The positions which are fixed, but not eliminated.

public:
    /// Default constructor.
    Node(void)
    {
    }

    /// \brief Constructor.
    /// \arg \c k0 The parameter k_0 of the type of the matrix \a mat.
    /// \arg \c mat A generator matrix.
    /// \arg \c blocks Maybe we can allready partition the columns. The canonization is done for the canonical Young subgroup
    /// defined by this ordered partition.
    Node(unsigned int k0, const Matrix<RING>& mat, const OrdPartition* const blocks = NULL) :
        m_matrix(mat), m_innerStab(k0, m_matrix.getCols()), m_eliminatedPos(m_matrix.getCols(), false)
    {
        if (blocks == NULL)
        {
            m_blocks = OrdPartition(mat.getCols());
        }
        else
        {
            m_blocks = *blocks;
        }
        HashGenerator::resize(std::max(m_matrix.getRows(), m_matrix.getCols() * RING::my_size() + 2));
    }

    /// Permute the columns of the matrix.
    void permuteMatrix(const Permutation& perm, const Permutation* from_root);

    /// \brief Calculate an invariant subset of the codewords.
    ///
    /// The invarian set should have at most MAXIMAL_INIT_WORKSIZE codewords.
    /// \todo We can use Magma for this job (but it is deactivated), but what should we do if we only have GAP?
    /// \todo Should we use other values for MAXIMAL_INIT_WORKSIZE or calculate it dynamically.
    bool initMinWords(AbstractCASModule*, const std::vector<unsigned int>& type);//, bool minWeightOfDual, unsigned int leastMinDist);
    bool initMinWordsByCAS(AbstractCASModule*, const std::vector<unsigned int>& type);//, bool minWeightOfDual, unsigned int leastMinDist);
    bool initMinWordsGrayCode(const std::vector<unsigned int>& type);//, bool minWeightOfDual, unsigned int leastMinDist);

    /// Get the information of fixed columns.
    const std::vector<bool>& get_isFixedCol(void) const
    {
        return m_innerStab.m_isFixedCol;
    }

    /// Get the matrix.
    const Matrix<RING>& getMatrix(void) const
    {
        return m_matrix;
    }

    /// Get the partition on the columns.
    const OrdPartition& getBlockSignature(void) const
    {
        return m_blocks;
    }

    /// Get the permutation which was applied to this node (on the path from the root node).
    const Permutation& getPermutation(void) const
    {
        return m_perm_to_sol;
    }

    /// How big is the invariant set of codewords we are working on the actual level of the tree (not the size of the initial set).
    unsigned int getMinWord_workSize(void) const
    {
        return m_minWord_workSize;
    }

    /// Which positions where used to shorten the invariant subset of codewords.
    const std::vector<bool>& getEliminatedPos(void) const
    {
        return m_eliminatedPos;
    }

    const std::vector<unsigned int>& getfixedUnEliminated(void) const
    {
        return m_fixedUnEliminated;
    }

    /// Get the inner stabilizer group.
    const InnerStab<RING, ALGTYPE>& getInnerStab(void) const
    {
        return m_innerStab;
    }

    /// Get an entry of the matrix.
    const Element<RING>& getEntry(int i, int j) const
    {
        return m_matrix.getElement(i, j);
    }

    /// Get the number of rows.
    unsigned int getRows(void) const
    {
        return m_matrix.getRows();
    }

    /// Get the number of columns.
    unsigned int getCols(void) const
    {
        return m_matrix.getCols();
    }

    /// Get the index of the column, which was fixed at level \a lvl.
    unsigned int getFixedCol(int lvl) const
    {
        return m_innerStab.m_fixedCols[lvl];
    }

    /// Go back in the backtrack tree to level  \a lvl, restore the state with the help of the arguments.
    inline void
    backTo(int lvl, unsigned int minWord_workSize, const std::vector<bool>& elim_pos,
            const std::vector<unsigned int> fixedUnEliminated, const OrdPartition& blocks, const InnerStab<RING,
                    ALGTYPE>& innerStab);

    /// \brief Calculate the information which block has to be fixed next.
    ///
    /// Usually we will take the smallest possible. But sometimes, this is not allowed because this would violate the reduced row echelon form.
    NextFixInfo calcNextFixInfo(const std::vector<unsigned int>& heights) const;

    /// \brief Is there a discrete block (size=1).
    ///
    /// If \a need_unit is true, we are not allowed to search for discrete blocks which leads to a pivot element that is nilpotent.
    bool hasDiscreteBlock(const std::vector<unsigned int>& heights, bool need_unit) const;

    /// Start the refinement methodes for the node.
    bool refine(int lvl, RefinementValues& best, bool& isBestInitialized, unsigned int k0,
            const GenNode<RING, ALGTYPE>* const p_backtrack);

    /// \brief Permute the columns such that \a image is minimal (under the canonical Young subgroup defined by \a m_blocks) and compare
    /// the images with the corresponding image for the candidate (stored in \a best).
    /// \returns <is possibly canonical, was a block refined>
    template<class T>
    std::pair<bool, bool> oneRefinement(std::vector<T>& image, std::vector<T>& best, bool& isBestInitialized, bool only_automs);

    /// Calculate a signature using the invariant set given by { m_minWords[0], ..., m_minWords[m_minWord_workSize-1] }
    /// todo: We can get a better invariant if we not only store m_perm_to_sol but also the applied column multiplications and the field automorphism.
    /// in this case, we could use the entries of the codewords at the positions m_fixedUnEliminated (up to normalization)
    //    std::vector<uchar>& minWordRefine(bool remove, std::vector<uchar>& res);

    /// Fix the column with index \a index.
    void fixCol(unsigned int index);

    /// Minimize the last fixed column under the inner stabilizer.
    void minimize(unsigned int k0, bool isInSpan, ActingElement<RING, ALGTYPE>* const g = NULL)
    {
        m_innerStab.minimize(k0, isInSpan, m_matrix, g);
    }

    /// Save the information of this node in \a bestCand and \a innerStab2BestCand.
    void saveAsBest(Matrix<RING>& bestCand, InnerStab<RING, ALGTYPE>& innerStab2BestCand) const;

    /// Write the information of the node in latex style.
    void toLatex(std::ostream& os) const;

    /// Output.
    template<typename R, AlgTypeE ATYPE>
    friend std::ostream & operator<<(std::ostream & os, const Node<R, ATYPE>& node);

    friend class RefineMethods<ALGTYPE> ;
    //
    //    template<typename R>
    //    friend std::vector<uchar>& minWordRefineChainRing4Linear(Node<R, LINEAR>& node, bool remove, std::vector<uchar>& res);
};

template<typename RING, AlgTypeE ALGTYPE>
void Node<RING, ALGTYPE>::backTo(int lvl, unsigned int minWord_workSize, const std::vector<bool>& elim_pos,
        const std::vector<unsigned int> fixedUnEliminated, const OrdPartition& blocks,
        const InnerStab<RING, ALGTYPE>& innerStab)
{
    m_innerStab = innerStab;
    m_blocks = blocks;
    m_minWord_workSize = minWord_workSize;
    m_eliminatedPos = elim_pos;
    m_fixedUnEliminated = fixedUnEliminated;
}

template<typename RING, AlgTypeE ALGTYPE>
void Node<RING, ALGTYPE>::permuteMatrix(const Permutation& perm, const Permutation* from_root)
{
    if (!perm.isId())
    {
        m_matrix.permuteCols(perm);
        if (from_root == NULL)
        {
            Permutation::mul(perm, m_perm_to_sol, m_perm_to_sol);
        }
        else
        {
            m_perm_to_sol = *from_root;
        }
    }
}

template<typename RING, AlgTypeE ALGTYPE>
bool Node<RING, ALGTYPE>::initMinWords(AbstractCASModule* cas, const std::vector<unsigned int>& type)
{
   return initMinWordsGrayCode(type);
//   unsigned int size1(m_minWords.size());
//   help::print(std::cout,m_minWords, "\n")<<std::endl;
//
//   initMinWordsByCAS(cas, type);
//   unsigned int size2(m_minWords.size());
//   if(size1 != size2)
//   {
//       std::cout<<"Fehler in MinWordBerechnung"<<std::endl;
//       help::print(std::cout,m_minWords, "\n")<<std::endl;
//
//       exit(-1);
//   }
//   return true;
}


template<typename RING, AlgTypeE ALGTYPE>
bool Node<RING, ALGTYPE>::initMinWordsByCAS(AbstractCASModule* cas, const std::vector<unsigned int>& type)
{
    ((MagmaCASModule*) cas)->getMinWords(m_matrix, m_minWords);
    m_minWord_workSize = m_minWords.size();
    return true;
}

template<typename RING, AlgTypeE ALGTYPE>
bool Node<RING, ALGTYPE>::initMinWordsGrayCode(const std::vector<unsigned int>& type)//, bool minWeightOfDual, unsigned int leastMinDist)
{
    m_minWords.clear();
    m_minWord_workSize = 0;
    typename RING::MyWeight minWeight(RING::MyWeight::minNonzero()); // smallest possible weight of a word
    Matrix<RING> information_span(m_matrix.getRows(), 0);
    unsigned int rank(0);

    do
    {
//        std::cout<<"Calculating invariant subset of codewords..."<<std::endl;
        CodewordEnumerator<RING> enumerator(type, m_matrix);
        typename RING::MyWeight actWeight(RING::MyWeight::max(m_matrix.getCols()));
        std::vector<DenseVec<RING> > newWords;
        std::vector<Vec<RING> > informationWords;

        while (++enumerator)
        {
            typename RING::MyWeight w(enumerator->getWeight());
            if (minWeight<=w)// && enumerator.getCodeword()[0]==1)
            {
                if(actWeight<w)
                {
                    continue;
                }
                if(w<actWeight)
                {
                    newWords.clear();
                    informationWords.clear();
                    actWeight = w;
                }
                if (enumerator.isProjPointRepresentative())// only one representative for each 1-dim subspace
                {
                    newWords.push_back( enumerator.getCodeword()  );
                    informationWords.push_back(enumerator.getInformationWord());
                }
            }
        }

        m_minWords.resize(m_minWord_workSize + newWords.size());
        std::copy(newWords.begin(), newWords.end() , m_minWords.begin()+m_minWord_workSize);
        m_minWord_workSize += newWords.size();

        information_span.appendColumns( informationWords );
        rank = information_span.ColumnSpan()[0];
        information_span.columnSubmatrix(rank);

//        std::cout<<"Words of small weights span a "<<rank<<" dimensional subspace."<<std::endl;
        minWeight = actWeight.successor();
    }while(rank != m_matrix.getRows());

    return true;
}


/*
template<typename RING, AlgTypeE ALGTYPE>
bool Node<RING, ALGTYPE>::initMinWordsGrayCode(const std::vector<unsigned int>& type)//, bool minWeightOfDual, unsigned int leastMinDist)
{
//    Distribution<SymWeight<1u> > wd;
    CodewordEnumerator<RING> enumerator(type, m_matrix);
    typename RING::MyWeight maxWeight = RING::MyWeight::max(m_matrix.getCols());
//    std::map<typename RING::MyWeight, std::vector<std::pair<Codeword<RING>, uchar> > > weight2Pairs;
    std::map<typename RING::MyWeight, std::vector<DenseVec<RING> > > weight2Words;
    unsigned int storedWords = 0;

//    typename CodewordEnumerator<RING>::CW lastCodeword ( *enumerator );

    while (++enumerator)
    {
        typename RING::MyWeight w(enumerator->getWeight());
        if (w < maxWeight)// && enumerator.getCodeword()[0]==1)
        {
            if (enumerator.isProjPointRepresentative())// only one representative for each 1-dim subspace
            {
               weight2Words[w].push_back( enumerator.getCodeword()  );
                ++storedWords;
                if (weight2Words.size() > 2 && (storedWords >= MAXIMAL_INIT_WORKSIZE || weight2Words.size()
                        > MAXIMAL_NR_OF_WEIGHTS))
                {
                    typename std::map<typename RING::MyWeight, std::vector<DenseVec<RING> > >::reverse_iterator rit(
                            weight2Words.rbegin());
                    storedWords -= rit->second.size();
                    maxWeight = rit->first;
                    weight2Words.erase(rit->first);
                }
            }
        }
    }

    m_minWords.resize(storedWords);
    typename std::map<typename RING::MyWeight, std::vector<DenseVec<RING> > >::iterator mapIter(weight2Words.begin());
    typename std::map<typename RING::MyWeight, std::vector<DenseVec<RING> > >::iterator mapIterEnd(weight2Words.end());
    typename std::vector<DenseVec<RING> >::iterator minWordIter(m_minWords.begin());
    for (; mapIter != mapIterEnd; ++mapIter)
    {
        std::copy(mapIter->second.begin(), mapIter->second.end() , minWordIter);
        minWordIter += mapIter->second.size();
    }

    m_minWord_workSize = m_minWords.size();
//    help::print(std::cout, m_minWords, "\n");
    return true;
}
*/

template<typename RING, AlgTypeE ALGTYPE>
bool Node<RING, ALGTYPE>::refine(int lvl, RefinementValues& best, bool& isBestInitialized, unsigned int k0,
        const GenNode<RING, ALGTYPE>* const p_backtrack)
{
    bool only_automs(p_backtrack->m_only_automs);
    unsigned int lastfixedPreimage(0);
    if (lvl != -1)
    {
        // we have to store  that the last fixed column is not used for a reduction of m_minWords
        lastfixedPreimage = m_perm_to_sol.invapply_neu(m_innerStab.m_fixedCols.back());
        m_fixedUnEliminated.push_back(lastfixedPreimage);
        //        Codeword<RING>::normalize(m_minWords, m_minWord_workSize, m_fixedUnEliminated); /// todo: How can we avoid this operation in each iteration
    }

    bool need_unit = m_innerStab.needUnit(k0);

    std::vector<unsigned int> heights;
    std::pair<bool, bool> oneRefineRes;

    if (m_matrix.getRows() != m_innerStab.m_rank)// this invariant is neccessary for the algorithm, must be always applied
    {
        heights = m_matrix.heightSequence(m_innerStab.m_rank);
        oneRefineRes = oneRefinement(heights, best.m_heights, isBestInitialized, only_automs);
        if (!oneRefineRes.first)
        {
            return false;
        }

        if (oneRefineRes.second)
        {
            if (p_backtrack->labraTest(m_perm_to_sol))
            {
                return false;
            }
        }
        if (hasDiscreteBlock(best.m_heights, need_unit)) // a discrete block
        {
            return true;
        }
    }
    else
    {
        best.m_heights = std::vector<unsigned int>(m_matrix.getCols(), RING::chainlength());
    }

    // the support can not change for columns in the span..., we additionally calculate some quotients
    std::vector<uchar> support;
    RefineMethods<ALGTYPE>::supportRefine(*this, best.m_heights, support);
    oneRefineRes = oneRefinement(support, best.m_support, isBestInitialized, only_automs);

    if (!oneRefineRes.first)
    {
        return false;
    }

    if (oneRefineRes.second)
    {
        if (p_backtrack->labraTest(m_perm_to_sol))
        {
            return false;
        }
    }

    if (hasDiscreteBlock(best.m_heights, need_unit)) // a discrete block
    {
        return true;
    }

    unsigned int counter = 0;
    //	cout<<*this<<std::endl;

    unsigned int oldminWordSize = m_minWord_workSize;
    bool forceSecondIteration = false;

    bool not_remove_flag=true;

    if (!not_remove_flag && lvl != -1)
    {
        // we have to store  the elimination based on this coordinate
        m_eliminatedPos[lastfixedPreimage] = true;
        m_fixedUnEliminated.pop_back();
    }

    while (true)
    {
        std::vector<DenseVec<RING> > actMinWords(m_minWords.begin(), m_minWords.begin()+m_minWord_workSize);
//        help::print(std::cout, actMinWords, "\n")<<std::endl;


        // we can use the signature based on the invariant set iteratively and as long as some blocks change.
        //		std::cout<<*this<<endl;
        std::vector<uchar> minWordImg;
        RefineMethods<ALGTYPE>::minWordRefine(*this, !not_remove_flag && (lvl != -1 && counter == 0), minWordImg);
//        help::print(std::cout, minWordImg, " ")<<std::endl;

        if (!isBestInitialized)
        {
            best.m_minWordImg.resize(counter + 1);
        }

        oneRefineRes = oneRefinement(minWordImg, best.m_minWordImg[counter], isBestInitialized, only_automs);

        if (!oneRefineRes.first)
        {
            return false;
        }

        if (m_minWord_workSize < oldminWordSize && m_minWord_workSize < MINIMAL_WORKSIZE)
        {
            // we removed too many words, we restore the old state.
            m_minWord_workSize = oldminWordSize;
            m_eliminatedPos[lastfixedPreimage] = false;
            m_fixedUnEliminated.push_back(lastfixedPreimage);
            forceSecondIteration = true;
        }

        if (oneRefineRes.second)
        {
            if (p_backtrack->labraTest(m_perm_to_sol))
            {
                return false;
            }
            if (hasDiscreteBlock(best.m_heights, need_unit)) // a discrete block
            {
                return true;
            }
        }
        else
        {
            if (!forceSecondIteration)
            {
                // the blocks did not change anymore
                return true;
            }
            else
            {
                forceSecondIteration = false;
            }
        }

        //		std::cout<<*this<<endl;
        ++counter;
    }
    return true;
}

template<typename RING, AlgTypeE ALGTYPE>
template<class T>
std::pair<bool, bool> Node<RING, ALGTYPE>::oneRefinement(std::vector<T>& image, std::vector<T>& best,
        bool& isBestInitialized, bool only_automs)
{
    ///\todo Maybe we can pass iterators and calculate the images on the flight.

    unsigned int blockstart = 0;
    std::vector<int> list_perm(m_matrix.getCols(), -1);
    bool didBlockSignature_change = false;

    if (!isBestInitialized)
    {
        best.resize(m_matrix.getCols());
    }

    while (blockstart != m_matrix.getCols())
    {
        if (m_innerStab.m_isFixedCol[blockstart])
        {
            // fixed blocks must be equal to the best candidate and provide no new information.
            ++blockstart;
            continue;
        }

        unsigned int blockend = blockstart + 1;
        while (blockend != m_matrix.getCols() && !m_blocks[blockend])
        {
            ++blockend;
        }
        // now we have a block of nonfixed columns, sort the images in the block and apply this permuation to the matrix
        // stop, if this leads to a nonoptima candidate
        Permutation g;
        g.stable_sorting_perm_inverse(image.begin() + blockstart, image.begin() + blockend);
        int deg = g.degree();
        T last_value(0);
        for (unsigned int j = 0; j < blockend - blockstart; ++j)
        {
            unsigned int preimage = g.apply(j, deg);
            list_perm[blockstart + j] = blockstart + preimage;

            T loc_value = image[blockstart + preimage];
            if (isBestInitialized)
            {
                T& valOfBest = best[blockstart + j];
                if(only_automs)
                {
                    if (loc_value != valOfBest)
                    {
                        return std::pair<bool, bool>(false, false);
                    }
                }
                else
                {
                    if (loc_value < valOfBest)
                    {
                        isBestInitialized = false;
                        valOfBest = loc_value;
                    }
                    if (loc_value > valOfBest)
                    {
                        return std::pair<bool, bool>(false, false);
                    }
                }
            }
            else
            {
                best[blockstart + j] = loc_value;
            }

            if (j == 0)
            {
                last_value = loc_value;
            }
            else if (loc_value != last_value)
            {
                didBlockSignature_change = true;
                m_blocks[blockstart + j] = true;
                last_value = loc_value;
            }
        }
        blockstart = blockend;
    }

    // apply the permutation to the matrix
    Permutation dummy, g;
    g.list(list_perm);
    Permutation::inv(g, dummy);
    permuteMatrix(dummy, NULL);

    return std::pair<bool, bool>(true, didBlockSignature_change);
}

template<>
template<>
std::vector<uchar>& RefineMethods<LINEAR>::minWordRefine<Z4>(Node<Z4, LINEAR>& node, bool remove,
        std::vector<uchar>& res);

template<>
template<>
std::vector<uchar>& RefineMethods<LINEAR>::minWordRefine<F2_F2>(Node<F2_F2, LINEAR>& node, bool remove, std::vector<
        uchar>& res);

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<Z4>(Node<Z4, PERMUTATIONAL>& node, bool remove,
        std::vector<uchar>& res);

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<F2_F2>(Node<F2_F2, PERMUTATIONAL>& node, bool remove,
        std::vector<uchar>& res);

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<BinaryField>(Node<BinaryField, PERMUTATIONAL>& node,
        bool remove, std::vector<uchar>& res);


template<>
std::vector<uchar>& RefineMethods<PLUSMINUSMONOMIAL>::supportRefine<Field>(const Node<Field, PLUSMINUSMONOMIAL>& node,
        const std::vector<unsigned int>& heights, std::vector<uchar>& res);

template<>
std::vector<uchar>& RefineMethods<PLUSMINUSMONOMIAL>::minWordRefine<Field>(Node<Field, PLUSMINUSMONOMIAL>& node, bool remove, std::vector<uchar>& res);

template<>
void RefineMethods<PLUSMINUSMONOMIAL>::addHashOfCodeword<Field>(const DenseVec<Field>& c,
        std::vector<uchar>::iterator& resIterBeg, std::vector<uchar>::iterator& resIterEnd, const std::vector<unsigned int>::const_iterator& posToBlockIterBeg,
        const std::vector<unsigned int>::const_iterator& posToBlockIterEnd, unsigned int blockcounter);

template<AlgTypeE ALGTYPE>
template<typename RING>
std::vector<uchar>& RefineMethods<ALGTYPE>::minWordRefine(Node<RING, ALGTYPE>& node, bool remove,
        std::vector<uchar>& res)
{
    /// be careful: must be "Field"

    unsigned int n(node.m_matrix.getCols());
    res = std::vector<uchar>(n, 0);
    std::vector<unsigned int> posToBlock(n);
    std::vector<unsigned int>::const_iterator posToBlockIterBeg(posToBlock.begin());
    std::vector<unsigned int>::const_iterator posToBlockIterEnd(posToBlock.end());
    Permutation inv_perm;
    Permutation::inv(node.m_perm_to_sol, inv_perm);
    int deg = inv_perm.degree(), blockcounter = 0;
    unsigned int lastFixedPreImg(0);
    if (remove)
    {
        lastFixedPreImg = inv_perm.apply(node.m_innerStab.m_fixedCols.back(), deg);
    }

    for (unsigned int col = 0; col < n; ++col)
    {
        if (col > 0 && node.m_blocks[col])
        {
            ++blockcounter;
        }
        posToBlock[inv_perm.apply(col, deg)] = blockcounter;
    }
    ++blockcounter;

    std::vector<unsigned int> blockSizes = node.m_blocks.getBlocksizes();
    std::vector<unsigned int>::const_iterator blockSizesBeg = blockSizes.begin();
    std::vector<unsigned int>::const_iterator blockSizesEnd = blockSizes.end();
    std::vector<DenseVec<Field> >::iterator wordIter(node.m_minWords.begin());
    std::vector<DenseVec<Field> >::reverse_iterator wordIterSwap(node.m_minWords.rbegin() + (node.m_minWords.size()
            - node.m_minWord_workSize));
    std::vector<bool>::const_iterator elimIterBeg(node.m_eliminatedPos.begin());

//    bool existsFixedUnEliminated(node.m_fixedUnEliminated.size() != 0);
//
//    std::vector<unsigned int>::const_iterator m_fixedUnElimIterBeg(node.m_fixedUnEliminated.begin());
//    std::vector<unsigned int>::const_iterator m_fixedUnElimIterEnd(node.m_fixedUnEliminated.end());

    for (unsigned int i = 0; i < node.m_minWord_workSize;)
    {
        /// If \a remove is set to true, we restrict ourselves to the subset of codewords, which are not zero on the last fixed position.
        if (remove && wordIter->operator[](lastFixedPreImg) == 0)
        {
            std::swap(*wordIter, *wordIterSwap);
            ++wordIterSwap;
            --node.m_minWord_workSize;
            continue;
        }

        std::vector<unsigned int> loc_res(blockcounter, 0);
        std::vector<bool>::const_iterator elimIter = elimIterBeg;
        Vec<Field>::const_iterator entryIter(wordIter->begin());
        std::vector<unsigned int>::const_iterator posToBlockIter(posToBlockIterBeg);

        for (; posToBlockIter < posToBlockIterEnd; ++entryIter, ++posToBlockIter, ++elimIter)
        {
            if (!*elimIter)
            {
                if (entryIter->isZero())
                {
                    ++loc_res[*posToBlockIter];
                }
            }
        }

        uchar myHash = HashGenerator::getHash(loc_res, blockSizesBeg, blockSizesEnd);

        // this is already done above...
//        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
//        if (existsFixedUnEliminated)
//        {
//            std::vector<unsigned int>::const_iterator m_fixedUnElimIter(m_fixedUnElimIterBeg);
//            for (; m_fixedUnElimIter < m_fixedUnElimIterEnd; ++m_fixedUnElimIter, ++hashIter)
//            {
//                if (wordIter->operator[](*m_fixedUnElimIter).isZero())
//                {
//                    myHash += *hashIter; // todo: if we would apply the column multiplication and the field automorphism, we can even use the value of the entry
//                }
//            }
//        }
        uchar zerohash(*HashGenerator::begin());

        for (unsigned int col = 0; col < n; ++col)
        {
            if (wordIter->operator[](col).isZero())
            {
                res[col] += myHash;
            }
            else
            {
                res[col] += myHash ^ zerohash;
            }
        }

        ++i;
        ++wordIter;
    }

    node.m_perm_to_sol.permute(res.begin(), res.end());
    return res;
}

template<>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::minWordRefine<Field>(Node<Field, PERMUTATIONAL>& node, bool remove,
        std::vector<uchar>& res);

template<>
void RefineMethods<PERMUTATIONAL>::addHashOfCodeword<Field>(const DenseVec<Field>& c,
        std::vector<uchar>::iterator& resIterBeg, std::vector<uchar>::iterator& resIterEnd,
        const std::vector<unsigned int>::const_iterator& posToBlockIterBeg,
        const std::vector<unsigned int>::const_iterator& posToBlockIterEnd, unsigned int blockcounter);

template<>
template<>
std::vector<uchar>& RefineMethods<SEMILINEAR>::supportRefine<Field>(const Node<Field, SEMILINEAR>& node,
        const std::vector<unsigned int>& heights, std::vector<uchar>& res);

template<typename RING>
std::vector<uchar>& RefineMethods<PERMUTATIONAL>::supportRefine(const Node<RING, PERMUTATIONAL>& node,
        const std::vector<unsigned int>& heights, std::vector<uchar>& res)
{
    unsigned int n = node.getCols();
    const std::vector<Vec<RING> >& matrixCols(node.getMatrix().getEntries());

    res = std::vector<uchar>(n, 0);
    std::vector<uchar>::iterator resIter(res.begin());
    std::vector<unsigned int>::const_iterator heightIter(heights.begin());
    typename std::vector<Vec<RING> >::const_iterator colIter(matrixCols.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(matrixCols.end());

    std::vector<bool>::const_iterator isFixIter(node.getInnerStab().m_isFixedCol.begin());
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
    unsigned int rank(node.getInnerStab().m_rank);

    for (; colIter < colIterEnd; ++colIter, ++resIter, ++heightIter, ++isFixIter)
    {
        if (*isFixIter)
        {
            continue;
        }
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (*heightIter == RING::chainlength())
        {
            typename std::vector<Element<RING> >::const_iterator colEntIter(colIter->begin());
            typename std::vector<Element<RING> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            unsigned int row = 0;

            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter, ++row)
            {
                *resIter += (*hashIter) * colEntIter->getIndex();
            }
        }
        else if (*heightIter == 1)
        {
            typename std::vector<Element<RING> >::const_iterator colEntIter(colIter->begin());
            typename std::vector<Element<RING> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter)
            {
                *resIter += (*hashIter) * (colEntIter->isUnit()); // an addition by 2 is still possible
            }
        }

    }
    return res;
}

template<>
template<>
std::vector<uchar>& RefineMethods<LINEAR>::supportRefine<Field>(const Node<Field, LINEAR>& node, const std::vector<
        unsigned int>& heights, std::vector<uchar>& res);

template<>
template<typename RING>
std::vector<uchar>& RefineMethods<LINEAR>::supportRefine(const Node<RING, LINEAR>& node,
        const std::vector<unsigned int>& heights, std::vector<uchar>& res)
{
    if (node.getInnerStab().m_type[1] == 0)
    {
        return supportRefineK0(node, heights, res);
    }
    else
    {
        return supportRefineK1(node, heights, res);
    }
}

template<AlgTypeE ALGTYPE>
template<typename RING>
std::vector<uchar>& RefineMethods<ALGTYPE>::supportRefineK0(const Node<RING, ALGTYPE>& node, const std::vector<
        unsigned int>& heights, std::vector<uchar>& res)
{
    // only on s0 component...
    unsigned int n(node.getCols());
    res = std::vector<uchar>(n, 0);
    std::vector<uchar>::iterator resIter(res.begin());
    std::vector<unsigned int>::const_iterator heightIter(heights.begin());
    const std::vector<Vec<RING> >& matrixCols(node.getMatrix().getEntries());
    typename std::vector<Vec<RING> >::const_iterator colIter(matrixCols.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(matrixCols.end());

    std::vector<bool>::const_iterator isFixIter(node.getInnerStab().m_isFixedCol.begin());
    std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());

    unsigned int rank(node.getInnerStab().m_rank);
    const UnordPartition& row_partition(node.getInnerStab().m_row_partition);

    for (; colIter < colIterEnd; ++colIter, ++resIter, ++heightIter, ++isFixIter)
    {
        if (*isFixIter)
        {
            continue;
        }
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (*heightIter == RING::chainlength())
        {
            std::vector<Element<RING> > quotient_crossed_rows(rank, 0);
            typename std::vector<Element<RING> >::const_iterator colEntIter(colIter->begin());
            typename std::vector<Element<RING> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            unsigned int row = 0;

            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter, ++row)
            {
                if (colEntIter->isUnit())
                {
                    int setIndex = row_partition.getSetIndex(row);
                    Element<RING>& quotient = quotient_crossed_rows[setIndex];
                    if (quotient.isZero())
                    {
                        quotient = colEntIter->multInverse();
                        *resIter += *hashIter;
                    }
                    else
                    {
                        *resIter += (*hashIter) * (quotient * (*colEntIter)).getIndex();
                    }
                }
                else
                {
                    *resIter += (*hashIter) * colEntIter->getIndex();
                }
            }
        }
        else if (*heightIter == 1)
        {// chainlength>1
            typename std::vector<Element<RING> >::const_iterator colEntIter(colIter->begin());
            typename std::vector<Element<RING> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter)
            {
                *resIter += (*hashIter) * (colEntIter->height() > 0); // an addition by 2 is still possible
            }
        }

    }
    return res;
}

template<AlgTypeE ALGTYPE>
template<typename RING>
std::vector<uchar>& RefineMethods<ALGTYPE>::supportRefineK1(const Node<RING, ALGTYPE>& node, const std::vector<
        unsigned int>& heights, std::vector<uchar>& res)
{
    unsigned int n(node.getCols());
    res = std::vector<uchar>(n, 0);
    std::vector<uchar>::iterator resIter(res.begin());
    std::vector<unsigned int>::const_iterator heightIter(heights.begin());
    const std::vector<Vec<RING> >& matrixCols(node.getMatrix().getEntries());
    typename std::vector<Vec<RING> >::const_iterator colIter(matrixCols.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(matrixCols.end());

    std::vector<bool>::const_iterator isFixIter(node.getInnerStab().m_isFixedCol.begin());

    unsigned int rank(node.getInnerStab().m_rank);
    unsigned int k0(node.getInnerStab().m_type[0]);
    bool innerStabMatrixRepSizeEq1(node.getInnerStab().m_matrixRep.size() == 1);

    for (; colIter < colIterEnd; ++colIter, ++resIter, ++heightIter)
    {
        if (*isFixIter)
        {
            continue;
        }
        std::vector<uchar>::const_iterator hashIter(HashGenerator::begin());
        if (*heightIter == 2)
        {
            typename std::vector<Element<RING> >::const_iterator colEntIter(colIter->begin());
            typename std::vector<Element<RING> >::const_iterator colEntIterEnd(colIter->begin() + k0);
            if (innerStabMatrixRepSizeEq1)
            {
                Element<RING> quotient(0);
                for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter)
                {
                    if (quotient.isZero())
                    {
                        if (colEntIter->isUnit())
                        {
                            quotient = *colEntIter;
                            *resIter += *hashIter;
                        }
                        else
                        {
                            *resIter += colEntIter->getIndex() * (*hashIter);
                        }
                    }
                    else
                    {
                        *resIter += (*colEntIter * quotient).getIndex() * (*hashIter);
                    }
                }
            }
            else
            {
                for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter)
                {
                    *resIter += (colEntIter->height() > 0) * (*hashIter); // an addition by 2 is still possible
                }
            }

            colEntIterEnd = colIter->begin() + rank;
            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter)
            {
                *resIter += colEntIter->height() * (*hashIter); // an addition on these positions is no more possible
            }
        }
        else //*heightIter == 1
        {
            typename std::vector<Element<RING> >::const_iterator colEntIter(colIter->begin());
            typename std::vector<Element<RING> >::const_iterator colEntIterEnd(colIter->begin() + rank);
            for (; colEntIter < colEntIterEnd; ++colEntIter, ++hashIter)
            {
                *resIter += (colEntIter->height() > 0) * (*hashIter); // an addition by 2 is still possible
            }
        }
    }
    return res;
}

template<typename RING, AlgTypeE ALGTYPE>
NextFixInfo Node<RING, ALGTYPE>::calcNextFixInfo(const std::vector<unsigned int>& heights) const
{
    unsigned int smallestHeight = RING::chainlength() - 1;
    unsigned int choosenHeight = smallestHeight;
    unsigned int blockstart = 0;
    unsigned int bestIndex(0), bestBlockSize(m_matrix.getCols() + 1);
    while (blockstart != m_matrix.getCols())
    {
        if (m_innerStab.m_isFixedCol[blockstart])
        {
            ++blockstart;
            continue;
        }

        unsigned int blockend = blockstart + 1;
        while (blockend != m_matrix.getCols() && !m_blocks[blockend])
        {
            ++blockend;
        }

        unsigned int blockHeight = heights[blockstart];
        unsigned int blocksize = blockend - blockstart;
        if (blocksize == 1 && (blockHeight == 0 || blockHeight == RING::chainlength()))
        {
            // a trivial block is nice
            return NextFixInfo(blockstart, blocksize, blockHeight == RING::chainlength());
        }

        if (blockHeight < smallestHeight) // we must take the smallest height or a block in the span...
        {
            smallestHeight = blockHeight;
            if (choosenHeight != RING::chainlength() || bestBlockSize > blocksize)
            {
                choosenHeight = blockHeight;
                bestBlockSize = blocksize;
                bestIndex = blockstart;
            }
        }
        else if (blockHeight == RING::chainlength() || blockHeight == smallestHeight)
        {
            if (bestBlockSize > blocksize)
            {
                bestBlockSize = blocksize;
                bestIndex = blockstart;
                choosenHeight = blockHeight;
            }
        }
        blockstart = blockend;
    }
    return NextFixInfo(bestIndex, bestBlockSize, choosenHeight == RING::chainlength());
}

template<typename RING, AlgTypeE ALGTYPE>
bool Node<RING, ALGTYPE>::hasDiscreteBlock(const std::vector<unsigned int>& heights, bool need_unit) const
{
    std::vector<unsigned int>::const_iterator heightIter(heights.begin());
    std::vector<bool>::const_iterator fixedIter(m_innerStab.m_isFixedCol.begin());
    std::vector<int>::const_iterator blocksIter(m_blocks.begin());
    std::vector<int>::const_iterator blocksIterEnd(m_blocks.end());

    for (; blocksIter < blocksIterEnd; ++blocksIter, ++fixedIter, ++heightIter)
    {
        if (*fixedIter)
        {
            continue;
        }
        if (*blocksIter && (((blocksIter + 1) == blocksIterEnd) || *(blocksIter + 1)))
        {
            if (!need_unit || (*heightIter != 1))
            {
                return true;
            }
        }
    }
    return false;
}

template<typename RING, AlgTypeE ALGTYPE>
void Node<RING, ALGTYPE>::fixCol(unsigned int index)
{
    if ((unsigned int) (index + 1) < m_matrix.getCols())
    {
        m_blocks[index + 1] = true;
    }
    m_innerStab.m_fixedCols.push_back(index);
    m_innerStab.m_isFixedCol[index] = true;
}

template<typename RING, AlgTypeE ALGTYPE>
void Node<RING, ALGTYPE>::saveAsBest(Matrix<RING>& bestCand, InnerStab<RING, ALGTYPE>& innerStab2BestCand) const
{
    bestCand = m_matrix;
    innerStab2BestCand = m_innerStab;
}

template<typename RING, AlgTypeE ALGTYPE>
std::ostream & operator<<(std::ostream & os, const Node<RING, ALGTYPE> & node)
{
    int rows = node.m_matrix.getRows();
    int cols = node.m_matrix.getCols();
    for (int i = 0; i < rows; ++i)
    {
        for (int j = 0; j < cols; ++j)
        {
            if (node.m_blocks[j])
            {
                os << "| ";
            }

            os << node.m_matrix.getElement(i, j) << " ";
        }

        if (i < rows - 1)
        {
            os << std::endl;
        }
    }
    return os;
}

#endif /*NODE_H_*/
