/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  ring.hpp
///  \brief Abstract class, defininig the arithmetic in a chain ring with 4 elements.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef RING_H_
#define RING_H_

#include <cstdlib>
#include <iostream>
#include <string>
#include "ringelement.hpp"
#include "helpfunction.hpp"

class Ring
{
    // 2 is the nilpotent element
public:
    static unsigned int chainlength(void)
    {
        return 2;
    }

    static unsigned int bitsPerEntry(void)
    {
        return 2;
    }

    inline static My_Long_Int SubmoduleQty(const unsigned int* const type, unsigned int n);

    inline static INDEX thetapow(unsigned int i);
    inline static unsigned int my_groundfield_size(void);
    inline static unsigned int my_size();
    inline static unsigned int my_multiplicativeGroupSize(void);
    inline static INDEX random();
    inline static INDEX randomInvertible();

    // the same for both rings
    inline static bool isUnit(const INDEX a);
    inline static INDEX mult(const INDEX a, const INDEX b);
    inline static INDEX multInverse(const INDEX a); // returns 0 if nilpotent
    inline static INDEX div(const INDEX a, const INDEX b);
    inline static unsigned int height(const INDEX a);
    inline static bool isReducable(const INDEX a);
    inline static int compare(const INDEX a, const INDEX b);
};

My_Long_Int Ring::SubmoduleQty(const unsigned int* const type, unsigned int n)
{
    unsigned int k = type[0] + type[1];
    if (k > n)
    {
        return 0;
    }

    My_Long_Int nec_qty = ((My_Long_Int) 1) << (type[0] * (n - k));
    nec_qty *= help::GaussBinomialBinary(n - type[0], type[1]);
    nec_qty *= help::GaussBinomialBinary(n, type[0]);
    return nec_qty;
}

INDEX Ring::thetapow(unsigned int i)
{
    switch (i)
    {
    case 0:
        return 1;
    case 1:
        return 2;
    default:
        std::cerr << "An error occured in Ring::thetapow" << std::endl;
        exit(-1);
    }
    return 4; // should not happen
}

unsigned int Ring::my_groundfield_size(void)
{
    return 2;
}

unsigned int Ring::my_size(void)
{
    return 4;
}

unsigned int Ring::my_multiplicativeGroupSize(void)
{
    return 2;
}

INDEX Ring::random()
{
    return rand() & 3;
}

INDEX Ring::randomInvertible()
{
    return 1 + 2 * (rand() & 1);
}

bool Ring::isUnit(const INDEX a)
{
    return a & 1;
}

INDEX Ring::mult(const INDEX a, const INDEX b)
{
    return (a * b) & 3;
}

INDEX Ring::multInverse(const INDEX a)
{
    switch (a)
    {
    case 1:
        return 1;
    case 3:
        return 3;
    default:
        std::cout << "Nilpotent element has no inverse" << std::endl;
        exit(-1);
    }
}

INDEX Ring::div(const INDEX a, const INDEX b)
{
    if ((a & 1) == 0)
    {
        std::cout << "Division by nilpotent element" << std::endl;
        exit(-1);
    }
    return mult(a, b); // b^(-1) == b
}

unsigned int Ring::height(const INDEX a)
{
    switch (a)
    {
    case 0:
        return 2;
    case 2:
        return 1;
    default:
        return 0;
    }
}

bool Ring::isReducable(const INDEX a)
{
    return a > 1;
}

// returns -1 if a<b, +1 if a>b and 0 if a=b
int Ring::compare(const INDEX a, const INDEX b)
{
    if (a == b)
    {
        return 0;
    }
    if(a==2)
    {
        return 1;
    }
    if(b==2)
    {
        return -1;
    }

    if(a<b)
    {
        return -1;
    }

    return 1; // a>b

//    unsigned int h_a(height(a)), h_b(height(b));
//    if (h_a < h_b)
//    {
//        return -1;
//    }
//    if (h_a > h_b)
//    {
//        return 1;
//    }
//    if (h_a == 0 && h_b == 0)
//    {
//        if (a < b)
//        {
//            return -1;
//        }
//        return 1;
//    }
    // should not happen
    std::cout << "there is an error in compare" << std::endl;
    exit(-1);
}

#endif /*RING_H_*/
